#include "hogwash.h"
/******************************************
* hogwash.c
*
* a simple libpcap packet filter
* designed to be used as an inline filter.
******************************************/

#include "control.h"
#include "pass_rule.h"
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <stdarg.h>

/*global variables*/
struct interface_rec	internal;
struct interface_rec	external;
char*					internal_name;
char*					external_name;

struct SaveRec*			int_head;
struct SaveRec*			ext_head;
struct SaveRec			int_recs[NUMBUFFS];
struct SaveRec			ext_recs[NUMBUFFS];
pthread_mutex_t			int_packet_mutex = PTHREAD_MUTEX_INITIALIZER;
pthread_mutex_t			ext_packet_mutex = PTHREAD_MUTEX_INITIALIZER;
PacketToProcess			int_packet;
PacketToProcess			ext_packet;

int						no_rules;
int						no_control;
FILE* 					stream;
int						debuglevel;
int						loglevel;

/*function prototypes*/
void InitStructures();
void InitInterface(struct interface_rec* r);
void CloseInterfaces();
void ShowBanner(int args);
int GetIfrMTU(char *name) ;
void ProcessPacket(char *user, struct pcap_pkthdr * pkthdr, u_char * pkt);
int OpenPcap(struct interface_rec* r, char* interface_name);
void CloseRawSock(struct interface_rec *r);
void* MainLoop(void* v);
void ParseArgs(int argc, char* argv[]);
void CreateLock();
void DestroyLock();
int hogwashprint (int loglevel, const char* format,...);

/****************************************************
* The one and only main
****************************************************/
int main(int argc, char* argv[]){

	/*save the args for when we restart*/
	GlobalArgs=argv;
	Restart=FALSE;
	MainPID=getpid();

	int_head=NULL;
	ext_head=NULL;

	/*testing*/
	hogwashprint(5,"binding signals\n");
	signal(SIGTERM, DestroyLock);
	signal(SIGINT,  DestroyLock);
	signal(SIGQUIT, DestroyLock);
	signal(SIGHUP,  HogwashRestart);
	/*end testing*/

	ShowBanner(FALSE);
	InitStructures();
	ParseArgs(argc, argv);
#ifdef DEBUG
	debuglevel=5;
#endif
	InitInterface(&internal);
	InitInterface(&external);
	if (!no_rules) pass_rule_init(internal.pds, argc, argv);
	if (!no_control) control_init(internal.pds, argc, argv);
	CreateLock();
	/*start up the handlers*/
	pthread_create(&external.thread, NULL, MainLoop, (void*)&external);
	pthread_create(&internal.thread, NULL, MainLoop, (void*)&internal);

	printf("Waiting for joins\n");
	pthread_join(external.thread, NULL);
	printf("external joined\n");
	pthread_join(internal.thread, NULL);
	printf("internal joined\n");
	
	
	if (Restart){
		printf("Threads Joined.  Restarting Now....\n");
		CloseInterfaces();
		pass_rule_free();
		remove(PID_FILE);
		printf("Calling execv\n");
		execv(argv[0], argv);
	}else{
		printf("About to exit for good\n");
		CloseInterfaces();
		pass_rule_free();
		DestroyLock();
		exit(0);
	}
	
	
}

/****************************************************
* Creates a PID so we don't run 2
****************************************************/
void CreateLock(){
	int			fp;
	struct stat	s;
	int			pid;
	int			statv;
	char		Buff[1024];
	
	bzero(&s, sizeof(struct stat));
	statv=stat(PID_FILE, &s);
	if (statv!=0){
		fp = open(PID_FILE, O_CREAT|O_RDWR);
		if (!fp){
			hogwashprint(5,"Couldn't create PID file %s.\n",PID_FILE);
			exit(1);
		}
		pid=getpid();
		sprintf(Buff,"%i\n",pid);
		write(fp, Buff, strlen(Buff));
		close(fp);
		return;
	}
	
	hogwashprint(5,"PID file %s exists. Running two copies of hogwash may crash some switches.\n", PID_FILE);  
	exit(1);
	
}

/***************************************************
* Erase the PID 
***************************************************/
void DestroyLock(){
	remove(PID_FILE);
	pass_rule_free();
	CloseInterfaces();
	exit(0);
}

/****************************************************
* The one and only ParseArgs :)
****************************************************/
void ParseArgs(int argc, char* argv[]){
	int	i=0;
	
	if (argc==1) ShowBanner(TRUE);
	
	for (i=0; i<argc; i++){	
		if (strncasecmp("-i",argv[i],2)==0){
			if ( (i>=argc-1) || (argv[i+1][0]=="-"[0])){
				hogwashprint(5,"Syntax -i <internal interface>\n");
				exit(1);
			}
			snprintf(internal_name, 10, argv[i+1]);
			hogwashprint(5,"New Internal %s\n",internal_name);
			i++;
		}else if (strncasecmp("-e",argv[i],2)==0){
			if ( (i>=argc-1) || (argv[i+1][0]=="-"[0]) ){
				hogwashprint(5,"Syntax -e <external interface>\n");
				exit(1);
			}
			snprintf(external_name, 10, argv[i+1]);
			hogwashprint(5,"New External %s\n",external_name);
		}else if (strncasecmp("-n", argv[i],2)==0){
			no_rules=TRUE;
			hogwashprint(5,"******************************************\n");
			hogwashprint(5,"* Entering No Rules Mode\n");
			hogwashprint(5,"******************************************\n");
		}
		else if (strncasecmp("-h",argv[i],2)==0)
		{
			ShowBanner(TRUE);
		}
		else if (strncasecmp("-k",argv[i],2)==0)
		{
			no_control=FALSE;
		}
		else if (strncasecmp("-d",argv[i],2)==0)
		{
			printf("Going into daemon mode\n");
			daemon(0,0);
		}

	}
}

/********************************************
* one for each thread
********************************************/
void* MainLoop(void* v){
	struct interface_rec*	r;
	
	printf("Starting Loop\n");
	r=(struct interface_rec*)v;
    if (r==&internal){
    	printf("Internal Interface %i\n", pthread_self());
		pthread_setcanceltype(PTHREAD_CANCEL_ASYNCHRONOUS, NULL);
		Thread1PID=getpid();
	}else{
		printf("External Interface %i\n", pthread_self());
		pthread_setcanceltype(PTHREAD_CANCEL_ASYNCHRONOUS, NULL);
		Thread2PID=getpid();
	}	
	
	pcap_loop(r->pds, -1, (pcap_handler) ProcessPacket, (char*)r->num);
	return NULL;
}

/***************************************
* print the banner so we know who we are
***************************************/
void ShowBanner(int args){
	printf("hogwash v0.02 by Jason Larsen\n");
	printf("using snort v1.8.1 by Martin Roesch\n\n");
	if (args){
		printf("Usage hogwash <options>\n");
		printf("  -c <rules file>\n");
		printf("  -C <rules file directory> (use highest numbered rules.vXXXXX file)\n");
		printf("  -i <internal interface>\n");
		printf("  -e <external interface>\n");
		printf("  -l <log directory>\n");
		printf("  -n (no rules mode)\n");
		printf("  -h (show this help message and exit)\n");
		printf("  -d (go to daemon mode)\n");
		printf("  -k <control config file>\n");
		exit(1);
	}
}

/**************************************
* set all variables to default values
**************************************/
void InitStructures(){
	no_rules=FALSE;
	no_control=TRUE;
	
	bzero(int_recs, sizeof(struct SaveRec) * NUMBUFFS);
	bzero(int_recs, sizeof(struct SaveRec) * NUMBUFFS);
	bzero(&int_packet, sizeof(PacketToProcess));
	bzero(&ext_packet, sizeof(PacketToProcess));
	
#ifdef DEBUG	
	stream=fopen (OUTPUT_FILE,"a");
#endif	
	
	internal_name=(char*)malloc(10);
	sprintf(internal_name, "eth0");
	external_name=(char*)malloc(10);
	sprintf(external_name, "eth1");

	internal.num=0;
	internal.string=internal_name;
	internal.pds=NULL;
	internal.snaplen = SNAPLEN;
	internal.mtu=0;
	internal.datalink=0;
	int_packet.interface=&internal;
	
	external.num=1;
	external.string=external_name;	
	external.pds=NULL;
	external.snaplen = SNAPLEN;
	external.mtu=0;
	external.datalink=0;
	ext_packet.interface=&external;
}

/**************************************
* startup the libpcap interfaces
**************************************/
void InitInterface(struct interface_rec* r){
	OpenPcap(r, r->string);
	OpenRawSock(r);
}

/***************************************
* Shut everything down
***************************************/
void CloseInterfaces(){
	CloseRawSock(&internal);
	CloseRawSock(&external);
	pcap_close(internal.pds);
	pcap_close(external.pds);
#ifdef DEBUG	
	fclose (stream);
#endif	
}

/*******************************************************
 * open the libpcap devices.  Mostly stolen from snort.
 *******************************************************/
int OpenPcap(struct interface_rec* r, char* interface_name){
	char errorbuf[1024];
    struct bpf_program fcode;    
	
	printf("In OpenPcap\n");
	if (!interface_name){
		r->string = pcap_lookupdev(errorbuf);
		hogwashprint(5,"found interface %s\n", r->string);
	}else{
		r->string=interface_name;
		hogwashprint(5,"Using guess %s\n", r->string);
	}
	r->pds = pcap_open_live(r->string, SNAPLEN,
         PROMISC , READ_TIMEOUT, errorbuf);
	 /* lookup mtu */
    r->mtu = GetIfrMTU(r->string);
	hogwashprint(5,"New MTU: %i\n",r->mtu);
	if (r->mtu == -1){
		printf("Couldn't get MTU assuming 1500\n");
		r->mtu=1500;
		//return FALSE;
	}
	
	if(pcap_compile(r->pds, &fcode, NULL, 1, 0) < 0)
    {
        hogwashprint(3,"ERROR: OpenPcap() FSM compilation failed: \n\t%s\n",
                     pcap_geterr(r->pds));
        hogwashprint(3,"PCAP command: %s\n", NULL);
    }
	
    /* set the pcap filter */
    if(pcap_setfilter(r->pds, &fcode) < 0)
    {
        hogwashprint(3,"ERROR: OpenPcap() setfilter: \n\t%s\n",
                   pcap_geterr(r->pds));
    }
	
    /* get data link type */
    r->datalink = pcap_datalink(r->pds);

    if(r->datalink < 0)
    {
        hogwashprint(3,"ERROR: OpenPcap() datalink grab: \n\t%s\n",
                   pcap_geterr(r->pds));
    }
	
	return TRUE;
}

/****************************************************************************
 *
 * Function  : GetIfrMTU()
 * Purpose   : Get Interface MTU value
 * Arguments : interface name (string)
 * Returns   : MTU (or -1)
 *
 ****************************************************************************/


int GetIfrMTU(char *name) {
    
    int fd;
    struct ifreq ifr;
    int retval;


    retval = -1;
    
    fd = socket(AF_INET, SOCK_DGRAM, 0);
    if ( fd < 0) {
        hogwashprint(1,"socket");
        return -1;
    }

    strncpy(ifr.ifr_name, name, sizeof(ifr.ifr_name));

#ifndef OSF1
    if (ioctl(fd, SIOCGIFMTU, &ifr) == 0)
        retval = ifr.ifr_metric;
#else
    if (ioctl(fd, SIOCRIPMTU, &ifr) == 0)
        retval = ifr.ifr_metric;
#endif
    else
        printf("ioctl(SIOCGIFMTU)");    

    close(fd);
    
 return retval;
}


/***********************************************
* Push a packet onto the queue
***********************************************/
void PushPacket(char* packet, int len, int which, struct timeval* tv){
	struct SaveRec*	seek;
	int				i;
	
	
	if (which==0){
		seek=&int_recs[0];
	}else{
		seek=&ext_recs[0];
	}
	
	/*find an open slot*/
	for (i=0;i<NUMBUFFS;i++){
		if (seek[i].len==0){
			memcpy(seek[i].Buff, packet, len);
			seek[i].len=len;
			memcpy(&seek[i].timestamp, tv, sizeof(struct timeval));
#ifdef DEBUG
			printf("Pushed into slot %i with len %i\n",i, seek[i].len);
#endif			
			return;
		}
		/*expire the buffer*/
		if ((tv->tv_sec-seek[i].timestamp.tv_sec) > BUFFTIMEOUT){
			seek[i].len=0;
			i--;
		}
	}
}




/*********************************************
* Pop a packet off the queue
* return FALSE if not found
*********************************************/
int PopPacket(char* packet, int len, int which){
	struct SaveRec* seek;
	int				i;
	
	struct SaveRec*	del;
	struct SaveRec*	head;
	struct SaveRec* last;
	
	if (which==0){
		seek=ext_recs;
	}else{
		seek=int_recs;
	}
	
	for (i=0;i<NUMBUFFS;i++){
		if (seek[i].len==len)
		if (memcmp(seek[i].Buff, packet, len)==0){
			/*found it.  delete and bail*/
			seek[i].len=0;
#ifdef DEBUG
			printf("Popped out of slot %i\n",i);
#endif						
			return TRUE;
		}
	}
	
	return FALSE;
}


void PrintPacket(unsigned char* p, int len){
	int i;
	
	for(i=0;i<len; i++){
		hogwashprint(2,"%02x", p[i]);
	}
	hogwashprint(2,"\n");
}


/*******************************************
* Each packet hits here
*******************************************/
void ProcessPacket(char *user, struct pcap_pkthdr * pkthdr, u_char * pkt){
	int						which;
	int						c;
	struct interface_rec*	r;
	PacketToProcess*		ptp;
	int						result;
	
	which=(int)user;
#ifdef DEBUG
	printf("Got packet on interface %i with len %i\n",which, pkthdr->len);
#endif
#ifdef DEBUGPATH
	printf("***************************\n");
	printf("Got packet on interface %i\n",which);
#endif

	/*see if this packet is a duplicate*/	
	if (which==0){
		/*came in on the internal interface*/
		r=&external;
		ptp=&ext_packet;
		ext_packet.flags=0;
		ext_packet.user=user;
		ext_packet.pkthdr=pkthdr;
		ext_packet.pkt=pkt;
		ext_packet.BuffLen=0;
		
		/*compare the packet for duplicates*/
		pthread_mutex_lock(&ext_packet_mutex);
		if (PopPacket(pkt, pkthdr->len, which)){
#ifdef DEBUGPATH
			printf("Packet was a duplicate\n");
#endif
			/*we are seeing a packet we just sent*/
			pthread_mutex_unlock(&ext_packet_mutex);
			return;
		}
		pthread_mutex_unlock(&ext_packet_mutex);
	}else{
		/*came in on the external interface*/
		r=&internal;
		ptp=&int_packet;
		int_packet.flags=0;
		int_packet.user=user;
		int_packet.pkthdr=pkthdr;
		int_packet.pkt=pkt;
		int_packet.BuffLen=0;
	
		
		pthread_mutex_lock(&int_packet_mutex);
		if (PopPacket(pkt, pkthdr->len, which)){
#ifdef DEBUGPATH
			printf("Packet was a duplicate\n");
#endif
			/*we are seeing a packet we just sent*/
			pthread_mutex_unlock(&int_packet_mutex);
			return;
		}		
		pthread_mutex_unlock(&int_packet_mutex);
	}

#ifdef DEBUG
	printf("Checking to see if it's a control packet\n");
#endif		
	if (!no_control){
		/*see if we drop this packet*/
		result = control_func(ptp);
		switch (result){
		case RESPONSE_DROP:
#ifdef DEBUG
			printf("Control Channel dropped the packet\n");
#endif		
			return;
		case RESPONSE_SEND_ON:
		case RESPONSE_SEND_ON | RESPONSE_USE_BUFFER:
			break;
		case RESPONSE_SEND_REPLY:
		case RESPONSE_SEND_REPLY | RESPONSE_USE_BUFFER:
			/*swap interfaces to send it out the correct one*/
			which= !which;
			if (which==0){
				r=&external;
			}else{
				r=&internal;
			}
			break;
		default:
			printf("Invalid Resonse Code\n");
			return;	
		}
#ifdef DEBUG
	if (result==RESPONSE_SEND_ON)
		printf("Wasn't a control Packet\n");
#endif
#ifdef DEBUGPATH
	printf("Passed the Control Channel\n");
#endif	
	}
#ifdef DEBUGPATH
	else{
		printf("No Control Channel\n");
	}
#endif
	
	/*if it passes the last check, send it on to the next one*/
	if (!no_rules){
		result=pass_rule(ptp);
		switch (result){
			case RESPONSE_DROP:
#ifdef DEBUGPATH
				printf("Pass Rule Dropped it\n");
#endif
				return;
			case RESPONSE_SEND_ON:
			case RESPONSE_SEND_ON | RESPONSE_USE_BUFFER:
#ifdef DEBUGPATH
				printf("Pass Rule Sent it on\n");
#endif
				break;
			case RESPONSE_SEND_REPLY:
			case RESPONSE_SEND_REPLY | RESPONSE_USE_BUFFER:
#ifdef DEBUGPATH
				printf("Pass Rule sent a reply\n");
#endif
				/*swap interfaces to send it out the correct one*/
				which= !which;
				if (which==0){
					r=&external;
				}else{
					r=&internal;
				}
				break;
			default:
				printf("Invalid Resonse Code\n");
				return;
		}
	}
#ifdef DEBUGPATH
	else{
		printf("Rules Not Enabled\n");
	}
#endif	

#ifdef DEBUG
	printf("Packet was passed with response code %i\n",result);
#endif	
	/*make a copy to check for duplicates*/
	if (which==0){
		pthread_mutex_lock(&int_packet_mutex);
		if (result & RESPONSE_USE_BUFFER){
			PushPacket(ptp->Buffer, ptp->BuffLen, which, &pkthdr->ts);
		}else{
			PushPacket(pkt, pkthdr->len, which, &pkthdr->ts);
		}	
		pthread_mutex_unlock(&int_packet_mutex);
	}else{
		pthread_mutex_lock(&ext_packet_mutex);
		if (result & RESPONSE_USE_BUFFER){
			PushPacket(ptp->Buffer, ptp->BuffLen, which, &pkthdr->ts);
		}else{
			PushPacket(pkt, pkthdr->len, which, &pkthdr->ts);
		}
		pthread_mutex_unlock(&ext_packet_mutex);
	}
#ifdef DEBUGPATH
	printf("Sending it on\n");
#endif
	
	/*send the packet out the other interface*/
	if (result & RESPONSE_USE_BUFFER){
		c = libnet_write_link_layer(r->network, r->string, ptp->Buffer, ptp->BuffLen); 
	}else{
		c = libnet_write_link_layer(r->network, r->string, pkt, pkthdr->len); 
	}
    if (c < pkthdr->len) 
    { 
        libnet_error(LN_ERR_WARNING, "libnet_write_link_layer only wrote %d bytes\n", c); 
    } 

}


/******************************************************
* creates a socket for writing
******************************************************/
int OpenRawSock(struct interface_rec* r){
    char 					err_buf[LIBNET_ERRBUF_SIZE]; 
		
	r->max_packet=1600;  /*ethernet*/
   	
	if ((r->network = libnet_open_link_interface(r->string, err_buf)) == NULL) 
    { 
        libnet_error(LIBNET_ERR_FATAL, "libnet_open_link_interface: %s\n", err_buf); 
    } 
	
	/*malloc max packet size*/
	 if (libnet_init_packet(r->max_packet, &r->packet) == -1) 
    { 
        libnet_error(LIBNET_ERR_FATAL, "libnet_init_packet failed\n"); 
    } 
 
	return TRUE;
}

/***********************************************
* Shut down the libnet interface
***********************************************/
void CloseRawSock(struct interface_rec *r){
    /* 
     *  Shut down the interface. 
     */ 
    if (libnet_close_link_interface(r->network) == -1) 
    { 
        libnet_error(LN_ERR_WARNING, "libnet_close_link_interface couldn't close the interface"); 
    } 
  

    /* 
     *  Free packet memory. 
     */ 
    libnet_destroy_packet(&r->packet); 
}

/* hogwashprint: print to either stdout or FILE* stream, depending on the
difference between loglevel and debuglevel.  If loglevel is less than or
equal to debuglevel, we vprintf to stdout; if it's greater than, we
vfprintf to stream.  We return the number of characters output */

int hogwashprint (int loglevel,const char* format,...)
	{
	va_list varargs;
	int file_or_stdout;
	va_start (varargs, format);
	file_or_stdout=loglevel-debuglevel;
	if (file_or_stdout >= 0 )
		{
		vprintf (format, varargs);
		}
	else
		{
		vfprintf (stream, format, varargs);
		}
		va_end (varargs);
	return (sizeof (format));
	}
	
	

/********************************************************
*  Reset and restart all the functions.  Reread rules
* and key files.
********************************************************/
void HogwashRestart(){
	printf("Restarting.......\n");
	printf("My PID   %i\n",getpid());
	printf("Main PID %i\n",MainPID);
	printf("T1 PID   %i\n",Thread1PID);
	printf("T2 PID   %i\n",Thread2PID);
	Restart=TRUE;
	printf("Stopping internal thread....\n");
	pthread_cancel(internal.thread);
	printf("Stopping external thread....\n");
	pthread_cancel(external.thread);
	printf("Done stopping threads\n");	
	
	/*send a HUP to the other thread*/
	
}
